; task.asm - *early* Multitasking support
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
;
; To do:
;	'pause' and 'break' functions. (Once we have user-mode programs, we
;	need wrapper functions to do syscalls)
;
;	What if we set DPL=3 for the IRQ0 descriptor?  Then any task can
;	call INT 0x20 to give up its timeslice.
;
;	Task variables to add:
;		input & output objects (standard buffers)
;		various flags
;
;	More flexible TSS allocation
;
; Notes:
;	Put TSS at the end of a page (at byte 0xF98), with the ioperm map on
;	the next page. That way, the ioperm can be changed just by changing
;	one entry in the page table.  The first byte (dword?) of the page
;	following the iomap must be 0xFF. (?)
;
;	386 manual, chap. 10.4.5: Before first task switch, need a valid TSS
;	and have TR point to a temporary area for the initial task state.
;		- done
;
;	Optimizations:
;		256-byte EVT_TBL alignment		- done
;

%include "boot.mac"
%include "memory.inc"
%include "macros.mac"

extern	mcopy
extern	putc, print, print_hex

%include "task.mac"

;%include "tss.inc"	;TSS structure
; Can't use STRUC because all the EQU's in it generate RDF import records..!
; Another reason to use XCOM :)

%define evt_size	(evt_end - evt_template)	;Event structure (header + TSS)
%define hdr_size	(tss_template - evt_template)	;Event header (before TSS)
%define tss_size	(evt_end - tss_template)	;TSS size (104+iomap)
%define tss_desc	(GDT+tss_sel)			;descriptor


	SEGMENT .text
init:	callproc print, msg_init
	call	task
	ret



	global task
task:
;
; Setup & start multitasking
;
	pushad

%if 0
callproc print_hex, new_eip
cr
mov eax, cr0
callproc print_hex, eax
cr
%endif

	;; Install timer_int on IRQ 0
	mov	eax, timer_int
	mov	[IDT+(0x20*8)+0], eax
	mov	[IDT+(0x20*8)+4], eax
	mov word [IDT+(0x20*8)+2], cs
	mov word [IDT+(0x20*8)+4], 0x8E00	;int gate, DPL=0

	;; Initialize scheduler tables
	mov	[EVT_TBL], dword 0	;Mark first one free

	mov	eax, esp
	callproc new_task,	new_eip, eax

%define tss0	EVT_TBL+256
	mov word [tss_desc], tss_size-1
	mov dword [tss_desc+2], tss0		;1st 3 bytes of addr
	mov word [tss_desc+5], 0x0089		;type=TSS, flags=Present
	mov byte [tss_desc+7], (tss0 >> 24)	;last byte

; callproc print_hex, [tss_desc+4]
; callproc print_hex, [tss_desc+0]
; cr

	;; load base & limit of outgoing (dummy) TSS
	mov	ax, tss_sel
	ltr	ax

	;; set up selector for incoming TSS
	mov byte [tss_desc+2], hdr_size	;Low byte of addr (same for all tasks)
	mov byte [tss_desc+5], 0x89	;Mark 'not busy'
	mov byte [tss_desc+7], EVT_TBL >> 24	;high byte (same for all tasks)


; Switch to the first task...
	mov word [tss_desc+3], (EVT_TBL >> 8) & 0xFFFF
	jmp	tss_sel:0		;switch


; Jump into the task-switch interrupt handler to start the first task...
;	mov	esi, EVT_TBL
;	jmp	dispatch_task


new_eip:  ;; Arrive here after switching to the first task
	mov	[EVT_TBL+256], dword 0	;quick hack to mark the slot "free" :)

	callproc set_timer, 500		;speed, in Hz
	in	al, 0x21		;enable timer IRQ
	and	al, 11111110b
	out	0x21, al

	popad
	ret


	global set_timer
set_timer:
	push	ebp
	mov	ebp, esp
	pusha
	mov	eax, 1191180	; calculate speed
	xor	edx, edx
	div dword [ebp+2*4]
	mov	ecx, eax
;	mov	al, 0x3A	; ch 0, mode 5, style 0
;	out	0x43, al	; (that's not what we need)
	mov	al, cl
	out	0x40, al
	mov	al, ch
	out	0x40, al
	popa
	leave
	ret


;; Some tests
	global t1, t2, t3, t4, t5
t1:	callproc new_task, .t, 0x70000
	ret
.t	jmp short $

t2:	callproc new_task, .t, 0x70100
	ret
.t	inc byte [0xb8010]
	jmp short .t

t3:	callproc new_task, .t, 0x70200
	ret
.t	inc byte [0xb8012]
	jmp short .t

t4:	callproc new_task, .t, 0x70300
	ret
.t	inc byte [0xb8014]
	jmp short .t

t5:	callproc new_task, .t, 0x70400
	ret
.t	inc byte [0xb8016]
	jmp short .t



	global new_task
new_task:
;
; Parameters:
%define .new_ip	ebp+3*4 ; EIP and ESP for new task
%define .new_sp	ebp+2*4
;
; For now, use the current CR3 (i.e., all tasks will share the same page
; tables).
;
	push	ebp
	mov	ebp, esp
	mpush	eax, esi, edi

	;Allocate space for TSS
	mov	esi, EVT_TBL-256
.alloc
	add	esi, 256
	cmp	[esi], dword 0
	jnz	.alloc

	;Create TSS
	callproc mcopy,  evt_template, esi, evt_size

	mov	eax, cr3	;Current CR3
	mov	[esi+hdr_size+28], eax

	mov	eax, [.new_ip]	;New EIP
	mov	[esi+hdr_size+32], eax

	pushfd			;Current EFLAGS
	pop	eax
	mov	[esi+hdr_size+36], eax

	mov	eax, [.new_sp]	;New ESP
	mov	[esi+hdr_size+56], eax

	;Link 'next' field to next task (or to self, if this is the only
	;task)
	cmp	esi, EVT_TBL
	je	.end
	mov	[esi-256], esi	;Link previous task to this one
	mov	[esi+256], dword 0	;Mark next spot free - quick hack :)
.end
	mov	[esi], dword EVT_TBL	;Link this task to task #0

	mpop	eax, esi, edi
	leave
	ret



timer_int:
	pusha
	irqack	0

	mov	eax, [current]		;Get current task
	mov	eax, [eax]		;Next
	mov	[current], eax

dispatch_task:
	shr	eax, 8			;256-byte alignment optimization
	mov	[tss_desc+3], ax	;modify TSS descriptor
	jmp	tss_sel:0		;switch

	popa
	iretd



	SECTION .data
msg_init	db	"++ Multitasking",0

;Scheduler variables
current	dd	EVT_TBL		;Current task

;Event variables (template)
evt_template:
	dd	EVT_TBL		;Next task in list
	dd	0		;flags

tss_template:
	dw	0,0		;0	backlink, 0
	dd	0,0, 0,0, 0,0	;4	esp0, esp1, esp2
	dd	0		;28	CR3
	dd	0,0		;32	EIP, EFLAGS
	dd	0,0,0,0		;40	eax, ecx, edx, ebx
	dd	0,0,0,0		;56	esp, ebp, esi, edi
	dw	0x10,0		;72	es
	dw	0x8,0		;76	cs
	dw	0x10,0		;80	ss
	dw	0x10,0		;84	ds
	dw	0x10,0		;88	fs
	dw	0x10,0		;92	gs
	dw	0,0		;96	LDTR, 0
	dw	0,0		;100	T, iobase
evt_end
