;; xcom.asm - XCOM loader

;; Notes:
;;	If we copy PROGRAM sections to a new location, do that before
;;	processing any other sections.
;;
;;	Is the anything in XCOM that can specify which routine(s) to execute
;;	immediately after loading?  Or do we just execute the first thing in
;;	.text, as with RDF?

	SECTION .text
LOAD_XCOM:	;; ( xcom -- next-xcom xt -1 | errflag )
;
; Input (tentative):
%define .object		ebp+4*4	;Address of object to load
%define .parent		ebp+3*4	;OID of object's parent-to-be
%define .flags		ebp+2*4	;flags to put in ODB entry
;
; Output (tentative):
%define .r_next		ebp+4*4	;Address at end of module (next object)
%define .code		ebp+3*4	;Address of .text segment
%define .result		ebp+2*4	;Result code
;
; Local Variables (remember to allocate enough space! SUB ESP, n*4)
%define .obj_len	ebp-1*4	;Length of module
%define .bss_len	ebp-2*4	;Length of BSS segment
%define .next		ebp-3*4	;Next object
%define .hdr_end	ebp-4*4	;End of RDF header
;
	mpush	eax,ebx,ecx,edx,esi,edi

	;; Check signature 'XCOM',19h,98h
	DPOP	esi
	lodsd
	cmp	eax, 'XCOM'
	jne	.badsig
	lodsw
	cmp	ax, 0x9819
	je	.ok

.badsig	callproc print,	.errmsg
	sprint "bad XCOM signature",10,0
	mov dword [.result], 2
	jmp	.end
.ok

	;; Read XCOM header
	lodsb
	mov	[refsize], al
	and byte [refsize], 0x1F
	shr	al, 7
	mov	[endian], al
	lodsb	;version - should be 1
	cmp	al, 1
	jbe	.h1
	mov	edi, 1	;error code
	jmp short .hdr_error
.1	lodsd
	mov	[filesize], eax
	lodsd
	mov	[sections], eax

	;; Read section headers
	mov	ecx, [sections]
.1	movzx	eax, byte [esi]		;type
pnum eax
	or	al, al			;check bounds (types 1-6 allowed)
	jz	.next
	cmp	al, 6
	ja	.next
	mpush	ecx, esi
	call	[SEC_HANDLERS + eax*4 - 4]
	mpop	ecx, esi
	mov	eax, [esi+8]		;print section name (if any)
	or	eax, eax
	jz	.next
	add	eax, [strings]
	space
	pstr	eax
.next	add	esi, 24			;next section header
	dcr
	loop	.1

.end	mpop	eax,ebx,ecx,edx,esi,edi
	ret
.hdr_error
	DPUSH	edi
	jmp short .end


SEC_HANDLERS:
	dd	sec_strings
	dd	sec_interface
	dd	sec_resource
	dd	sec_program
	dd	sec_symbols
	dd	sec_relocs

sec_strings:
	mov	eax, [esi+20]	;offset
	add	eax, [.object]
	mov	[strings], eax
	ret

sec_interface:		;ignore
sec_resource:		;ignore
	ret

sec_program:
%if 0
	;; Copy to new location
	mpush	esi
	mov	ecx, [esi+16]	;size
	mov	esi, [esi+20]	;offset
	add	esi, [.object]
	mov	edi, [DATA_end]	;new location
	mov	eax, edi
	cld
	rep	movsb
	mov	[DATA_end], edi	;store new location in section header
	mpop	esi
	sub	eax, [.object]
	mov	[esi+20], eax
%endif
	ret

sec_symbols:
	mov	ebx, [esi+12]	;PROGRAM section to operate on
	mov	ecx, [esi+16]	;size of SYMBOLS sections
	shr	ecx, 4		;# of records
	mov	esi, [esi+20]	;offset
	add	esi, [.object]
.l1
	test byte [esi+2], 2	;extern (import) ??
	jnz	.import

	;; Export
.export	; To do...

	jmp short .n

	;; Import
.import	; To do...
%if 0	;; First, get rid of counted strings, to make this simple...
	;;	name, len, parent, ifs=0
	mpush	esi, eax, [.parent], 0
	call	search
	pop	edx		;address
	add	esp, 3*4

	or	edx, edx
	jz	.error
%endif

.n	add	esi, 16		;next symbol record
	loop	.l1
	ret

sec_relocs:
	ret
eLOAD_XCOM


	SECTION .data


	SECTION .bss
filesize	resd	1
sections	resd	1	; number of sections
strings		resd	1	; ptr to string table
refsize		resb	1
endian		resb	1	; 0=little, 1=big
